(* The Denotational Semantics of Core delimcc *)


module type Delimcc = sig
  type 'a repr                          (* type of values *)
  type 'a res                           (* type of computation results *)


  (* values *)
  val int: int -> int repr
  val add: (int->int->int) repr
  val unit: unit repr

  type univ                             (* the universal type *)
  val i_univ: 'a repr -> univ repr
  val p_univ: univ repr -> 'a res

  val abs: ('a repr -> 'b res) -> ('a->'b) repr
  val absrec: (('a->'b) repr -> 'a repr -> 'b res) -> ('a->'b) repr

  type ('a,'b) free = 
    | Ret of univ repr 
    | Act of 'a repr * ('b -> ('a,'b) free) repr
  (* constructors: introduction *)
  val ret: univ repr -> ('a,'b) free repr
  val act: 'a repr -> ('b -> ('a,'b) free) repr -> ('a,'b) free repr
  (* destructor: elimination *)
  val with_free: ('a,'b) free repr -> 
                 (univ repr -> 'w res) ->
                 ('a repr -> ('b -> ('a,'b) free) repr -> 'w res) ->
                 'w res

  (* computations *)
  val vl: 'a repr -> 'a res                     (* all values are computations*)
  val let_: 'a res -> ('a repr -> 'b res) -> 'b res
  val ($$): ('a -> 'b) repr -> 'a repr -> 'b res

  (* The delimcc part: prompt and shift *)
  type 'a prompt
  val newpr: unit -> 'a prompt res
  val pushpr: 'a prompt repr -> 'a res -> 'a res
  val sh0:    'a prompt repr -> (('b -> 'a) repr -> 'a res) -> 'b res
end

(* The classical shift test: Danvy-Filinski Example
  (+ 10 (reset (+ 2 (shift k (+ 100 (k (k 3)))))))
  ; --> 117
  It also works with shift0

  Here is how the code looks in delimcc:
let test5 = 
  let p0 = new_prompt () in
  10 + push_prompt p0 @@ fun () -> 
   2 + shift0 p0 (fun sk -> 100 + sk (sk 3))
*)

module ExDF(D:Delimcc) = struct
 open D

 (* First: a bit of syntax sugar: D is a very bare-bone DSL
    It also distinguishes expressions and values, in types
 *)

  (* A macro to apply a computation: in contrast, ($$) applies a value *)
 let ($$>) v e = let_ e (fun x -> v $$ x)

 (* Add an expression to an integer value *)  
 let (++%) v e = let_ (add $$ v) (fun fv -> let_ e (fun ev -> fv $$ ev)) 

 let ans = 
  let_ (newpr ()) @@ fun p0 -> 
  int 10 ++% 
    pushpr p0 (int 2 ++% sh0 p0 (fun sk -> int 100 ++% (sk $$> (sk $$ int 3))))
end

(* A more challenging example exercising multiple prompts: 
   the variation of famous Danvy-Filinski Example, 
   test7'' from delimcc test suite *)
module ExD(D:Delimcc) = struct
 open D

  (* A macro to apply a computation: ($$) applies a value *)
 let ($$$) e x = let_ e (fun z -> z $$ x)

 let (++%) e v = let_ e (fun ev -> let_ (add $$ ev) (fun fv -> fv $$ v))

 let ans = 
  let_ (newpr ()) @@ fun p1 -> 
  let_ (newpr ()) @@ fun p2 -> 
  let_ (newpr ()) @@ fun p3 -> 
  let pushtwice sk = (* OCaml let: macro *)
    sk $$ abs (fun (_:unit repr) -> 
      sk $$ abs (fun (_:unit repr) ->
	sh0 p2 (fun sk2 -> sk2 $$ abs (fun (_:unit repr) -> 
           sk2 $$ abs (fun (_:unit repr) -> vl (int 3)))) $$$ unit)) in
  pushpr p1 (
    pushpr p2 (
      pushpr p3 (sh0 p1 pushtwice $$$ unit) ++% int 10) ++% int 1) ++% int 100
end
(* 135 *)

(*
module ExD :
  functor (D : Delimcc) ->
    sig
      val ( $$$ ) : ('a -> 'b) D.res -> 'a D.repr -> 'b D.res
      val ( ++% ) : int D.res -> int D.repr -> int D.res
      val ans : int D.res
    end
*)

(* The two expressions to demonstrate the shift0 rewriting rule *)

module Shift0T(D:Delimcc) = struct
 open D

 let exprs p arg e =

  let el = pushpr p (let_ (sh0 p (fun k -> e k)) (fun opv ->
                   let_ arg                  (fun argv -> 
                   opv $$ argv)))
  in
  let er = 
   let k = abs (fun x -> 
               pushpr p (
                let_ (vl x) (fun opv ->
                let_ arg    (fun argv -> 
                opv $$ argv))))
   in e k
  in (el,er)
end
;;

module RDelimcc = struct
  type 'a repr  = 
    | B : 'a -> 'a repr                         (* OCaml values *)
    | F : ('a repr->'b res) -> ('a->'b) repr    (* Functions V->R *)
  and _ res   =                                 (* Results *)
    | V: 'a repr -> 'a res                      (* Normal termination result *)
    | E: {prompt: int;                          (* Bubble *)
          body:('c ->'w) repr->'w res;
          k: 'c repr -> 'a res} -> 'a res  

  let rec lift : ('a repr -> 'b res) -> 'a res -> 'b res = fun f -> function
    | V v -> f v
    | E ({k;_} as oper) -> E {oper with k = fun c -> lift f (k c)}


  (* values *)
  let int (x:int) = B x
  let add : (int->int->int) repr = 
    F (function B x -> V (F (function B y -> V (B (x+y)))))
  let unit = B ()

  type univ = Obj.t                            (* the universal type *)
  let i_univ: 'a repr -> univ repr = fun x -> B (Obj.repr x)
  let p_univ: univ repr -> 'a res  = function B x -> V (Obj.obj x)

  let abs f = F f
  let absrec: (('a->'b) repr -> 'a repr -> 'b res) -> ('a->'b) repr = fun f ->
    abs (fun x -> let rec h y = f (abs h) y in h x)

  let vl v = V v                     (* all values are computations *)
  let let_: 'a res -> ('a repr -> 'b res) -> 'b res = fun e f -> 
    lift f e
  let ($$): ('a -> 'b) repr -> 'a repr -> 'b res = 
    function F f -> fun x -> f x

  type ('a,'b) free = 
    | Ret of univ repr 
    | Act of 'a repr * ('b -> ('a,'b) free) repr
  (* constructors: introduction *)
  let ret: univ repr -> ('a,'b) free repr = fun x -> B (Ret x)
  let act: 'a repr -> ('b -> ('a,'b) free) repr -> ('a,'b) free repr =
    fun v k -> B (Act (v,k))
  (* destructor: elimination *)
  let with_free: ('a,'b) free repr -> 
                 (univ repr -> 'w res) ->
                 ('a repr -> ('b -> ('a,'b) free) repr -> 'w res) ->
                 'w res =
   fun (B free) reth acth -> match free with
   | Ret x -> reth x
   | Act (a,k) -> acth a k

  type 'a prompt = int
  let newpr: unit -> 'a prompt res =
    let c = ref 0 in
    fun () -> incr c; V (B !c)

  let sh0: 'a prompt repr -> (('b -> 'a) repr -> 'a res) -> 'b res =
    fun (B prompt) body -> E {prompt;body;k=vl}

       (* The appearance of Obj.magic comes from the fact that we don't
          carry the answer type in the type of a computation. Therefore,
          the answer-type is existentialized
          in E.
        *)
  let rec pushpr: 'a prompt repr -> 'a res -> 'a res = fun (B p) -> function
    | V x -> V x
    | E{prompt; body;k} 
      when prompt = p ->
        (* When prompt = p then 'w is 'a *)
        let (body:_->_) = Obj.magic body in
        body (abs (fun c -> pushpr (B p) (k c)))
         (* Relay to an outer handler *)
    | E ({k;_} as oper) -> E {oper with k = fun c -> pushpr (B p) (k c)}
end
;;

let RDelimcc.V (RDelimcc.B 117) = let module M = ExDF(RDelimcc) in M.ans

let RDelimcc.V (RDelimcc.B 135) = let module M = ExD(RDelimcc) in M.ans
;;

let _ = print_endline "All Done";;

